# ruff: noqa: D400, D415
"""
_____________________________________________________________________

:PROJECT: LARAsuite

*lara_django_store celery tasks.*

:details: lara_django_store celery tasks.
"""

# from config import celery_app

# @celery_app.task()



# ruff: noqa: D400, D415
"""
_____________________________________________________________________

:PROJECT: LARAsuite

*lara_django_store urls.*

:details: lara_django_store urls module.
         - add app specific urls here
"""

# !! this sets the apps namespace to be used in the template
app_name = "lara_django_store"



"""
_____________________________________________________________________.

:PROJECT: LARAsuite

*lara_django_store views *

:details: lara_django_store views module.
         - add app specific urls here
         -
:authors: mark doerr <mark.doerr@uni-greifswald.de>

.. note:: -
.. todo:: -
________________________________________________________________________
"""
from typing import Any
from django.contrib.auth.mixins import LoginRequiredMixin
from django.http import HttpResponse
from django.shortcuts import get_object_or_404, redirect
from django.views.generic import CreateView, DeleteView, DetailView, TemplateView, UpdateView, View
from django_tables2 import RequestConfig


class BaseStoreView:
    """Base mixin for all material store views."""

    @property
    def item_type(self) -> str:
        """Override in subclasses: 'device', 'part', 'labware'."""
        raise NotImplementedError("Subclasses must define item_type")

    @property
    def model_class(self) -> type:
        """Override in subclasses to return the main model class."""
        raise NotImplementedError("Subclasses must define model_class")

    @property
    def basket_class(self) -> type:
        """Override in subclasses to return the basket model class."""
        raise NotImplementedError("Subclasses must define basket_class")

    @property
    def order_class(self) -> type:
        """Override in subclasses to return the order model class."""
        raise NotImplementedError("Subclasses must define order_class")

    @property
    def order_item_class(self) -> type:
        """Override in subclasses to return the order item model class."""
        raise NotImplementedError("Subclasses must define order_item_class")

    def get_base_context(self, **kwargs: Any) -> dict[str, Any]:
        """Get common context data for all views."""
        context: dict[str, Any] = super().get_context_data(**kwargs) if hasattr(super(), "get_context_data") else {}  # type: ignore[misc]
        context["menu_items"] = []  # StoreMenu.menu_items
        context["list_url"] = f"lara_django_material_store:{self.item_type}-list"
        context["create_url"] = f"lara_django_material_store:{self.item_type}-create"
        context["baskets_url"] = f"lara_django_material_store:{self.item_type}-baskets"
        context["orders_url"] = f"lara_django_material_store:{self.item_type}-orders"
        return context


class BaseListView(BaseStoreView, TemplateView):
    """Generic list view for all item types."""

    template_name = "lara_django_material_store/list.html"

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:  # noqa: D102
        context = self.get_base_context(**kwargs)
        context["section_title"] = f"{self.item_type.title()} Store - List"

        # Get baskets for current user
        context["baskets"] = self.basket_class.objects.filter(entity=self.request.user.entity).order_by("name_display")  # type: ignore[attr-defined]

        # Set current basket
        basket_id = self.kwargs.get("basket_id")
        if basket_id and self.basket_class.objects.filter(entity=self.request.user.entity, pk=basket_id).exists():  # type: ignore[attr-defined]
            context["current_basket"] = self.basket_class.objects.get(pk=basket_id)  # type: ignore[attr-defined]
        else:
            context["current_basket"] = (
                self.basket_class.objects.filter(entity=self.request.user.entity)  # type: ignore[attr-defined]
                .order_by("datetime_last_modified")
                .last()
            )

        # Get items and create table pairs
        items = self.get_items_for_list()
        context[f"{self.item_type}_table_pairs"] = self.create_table_pairs(items)

        return context

    def get_items_for_list(self) -> list[Any]:
        """Override in subclasses to define how to get items."""
        raise NotImplementedError("Subclasses must implement get_items_for_list")

    def create_table_pairs(self, items: Any) -> list[Any]:
        """Override in subclasses to create table pairs."""
        raise NotImplementedError("Subclasses must implement create_table_pairs")


class BaseDetailView(BaseStoreView, DetailView):
    """Generic detail view for all item types."""

    template_name = "lara_django_material_store/detail.html"

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:  # noqa: D102
        context = self.get_base_context(**kwargs)
        context["section_title"] = f"{self.item_type.title()} Store - Details"
        context["update_link"] = f"lara_django_material_store:{self.item_type}-update"
        return context


class BaseCreateView(BaseStoreView, LoginRequiredMixin, CreateView):
    """Generic create view for all item types."""

    template_name = "lara_django_material_store/create_form.html"

    def get_success_url(self) -> str:  # noqa: D102
        return f"/material-store/{self.item_type}/list"

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:  # noqa: D102
        context = self.get_base_context(**kwargs)
        context["section_title"] = f"{self.item_type.title()} Store - Create"
        return context


class BaseUpdateView(BaseStoreView, LoginRequiredMixin, UpdateView):
    """Generic update view for all item types."""

    template_name = "lara_django_material_store/update_form.html"

    def get_success_url(self) -> str:  # noqa: D102
        return f"/material-store/{self.item_type}/list"

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:  # noqa: D102
        context = self.get_base_context(**kwargs)
        context["section_title"] = f"{self.item_type.title()} Store - Update"
        context["delete_link"] = f"lara_django_material_store:{self.item_type}-delete"
        return context


class BaseDeleteView(BaseStoreView, LoginRequiredMixin, DeleteView):
    """Generic delete view for all item types."""

    template_name = "lara_django_material_store/delete_form.html"

    def get_success_url(self) -> str:  # noqa: D102
        return f"/material-store/{self.item_type}/list"

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:  # noqa: D102
        context = self.get_base_context(**kwargs)
        context["section_title"] = f"{self.item_type.title()} Store - Delete"
        return context


class BaseBasketsView(BaseStoreView, LoginRequiredMixin, TemplateView):
    """Generic baskets view for all item types."""

    template_name = "lara_django_material_store/baskets.html"

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:  # noqa: D102
        context = self.get_base_context(**kwargs)
        context["section_title"] = f"{self.item_type.title()} Store - Baskets"

        baskets = self.basket_class.objects.filter(entity=self.request.user.entity).prefetch_related(  # type: ignore[attr-defined]
            f"{self.item_type}orderitem_set"
        )

        if not baskets.exists():
            return context

        basket_table_pairs = []
        for basket in baskets:
            basket.save()
            table = self.get_basket_table_class()(getattr(basket, f"{self.item_type}orderitem_set").all())
            RequestConfig(self.request).configure(table)
            basket_table_pairs.append({"basket": basket, "table": table})

        context["basket_table_pairs"] = basket_table_pairs
        return context

    def get_basket_table_class(self) -> type:
        """Override in subclasses to return the appropriate table class."""
        raise NotImplementedError("Subclasses must implement get_basket_table_class")


class BaseAddToBasketView(BaseStoreView, LoginRequiredMixin, View):
    """Generic add to basket view for all item types."""

    def get(self, request: Any, pk: Any, basket_id: Any, *args: Any, **kwargs: Any) -> Any:  # noqa: D102
        item = get_object_or_404(self.model_class, pk=pk)
        basket = get_object_or_404(self.basket_class, entity=request.user.entity, pk=basket_id)
        basket.add_item_instance(item)  # This method needs to be standardized
        return redirect(f"lara_django_material_store:{self.item_type}-list")


class BaseBasketAdjustQuantityView(BaseStoreView, LoginRequiredMixin, TemplateView):
    """Generic quantity adjustment view for all item types."""

    def post(self, request: Any) -> HttpResponse:  # noqa: D102
        item_id = request.POST.get(f"{self.item_type}_order_item_id")
        action = request.POST.get("action")

        item = get_object_or_404(
            self.order_item_class, pk=item_id, **{f"{self.item_type}_basket__entity": request.user.entity}
        )
        basket = getattr(item, f"{self.item_type}_basket")

        if action == "increase":
            item.quantity += 1
        elif action == "decrease":
            item.quantity = max(1, item.quantity - 1)
        item.save()

        table = self.get_basket_table_class()(
            self.order_item_class.objects.filter(**{f"{self.item_type}_basket": basket})  # type: ignore[attr-defined]
        )
        RequestConfig(request).configure(table)
        table_html = table.as_html(request)

        response_html = f"""
            <div class="table-container">{table_html}</div>
            <span id='total-price-{basket.pk}' hx-swap-oob='true'>{basket.total_price_net}</span>
        """
        return HttpResponse(response_html)



# ruff: noqa: D400, D415
"""
_____________________________________________________________________

:PROJECT: LARAsuite

*lara_django_store admin.*

:details: lara_django_store admin module admin backend configuration.
:authors: mark doerr <mark.doerr@uni-greifswald.de>

.. note:: -
.. todo:: - run "lara-django-dev admin_generator lara_django_store >> admin.py" to update this file
________________________________________________________________________
"""

# -*- coding: utf-8 -*-
from django.contrib import admin

from .models import HazardClass, SafetyInformation


@admin.register(HazardClass)
class HazardClassAdmin(admin.ModelAdmin):
    """Admin interface for HazardClass model."""

    list_display = (
        "name",
        "iri",
        "description",
        "hazardclass_id",
    )


@admin.register(SafetyInformation)
class SafetyInformationAdmin(admin.ModelAdmin):
    """Admin interface for SafetyInformation model."""



"""
_____________________________________________________________________.

:PROJECT: LARAsuite

*lara_django_store Model Context Protocol (MCP) handler *

:details: lara_django_store Model Context Protocol (MCP) handler.

:authors: mark doerr <mark.doerr@uni-greifswald.de>

________________________________________________________________________
"""

from mcp_server import ModelQueryToolset

from .models import HazardClass, SafetyInformation


class HazardClassQueryTool(ModelQueryToolset):
    """MCP query tool exposing read/query operations for `HazardClass`."""

    model = HazardClass


class SafetyInformationQueryTool(ModelQueryToolset):
    """MCP query tool exposing read/query operations for `SafetyInformation`."""

    model = SafetyInformation



"""
_____________________________________________________________________.

:PROJECT: LARAsuite

*lara_django_store models *

:details: lara_django_store database models.
         -
:authors: mark doerr <mark.doerr@uni-greifswald.de>

________________________________________________________________________
"""

import datetime
import logging
import uuid
from typing import Any

from django.conf import settings
from django.contrib.postgres.indexes import GinIndex
from django.contrib.postgres.search import SearchVectorField
from django.core.exceptions import ValidationError
from django.db import models
from django.utils import timezone
from lara_django_base.models import Currency, ExternalResource, ItemStatus, Location, Namespace, Tag
from lara_django_people.models import Entity, Group


class HazardClass(models.Model):
    """
    (Bio-)Hazard Class Model for substances, polymers, mixtures, ... .

    implements the hazard classification  Globally Harmonized System of Classification and Labelling of Chemicals (GHS)
    Standard signal word, hazard classes, pictograms, ...
    """

    model_curi = "lara:store/HazardClass"
    hazardclass_id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name = models.TextField(unique=True, default="", help_text="name of the hazard class")
    signal_word = models.TextField(blank=True, default="", help_text="signal word of the hazard class")
    description = models.TextField(blank=True, default="", help_text="description of the hazard class")
    pictogram_svg = models.TextField(
        blank=True,
        help_text="pictogram of the hazard class",
    )
    pictogram = models.ImageField(
        upload_to="substances_store/images/",
        blank=True,
        null=True,
        help_text="pictogram of the hazard class",
    )
    iri = models.URLField(
        blank=True,
        null=True,
        unique=True,
        max_length=512,
        help_text="International Resource Identifier - IRI: is used for semantic representation ",
    )

    def __str__(self) -> str:
        """Return string representation."""
        return self.name or ""

    def __repr__(self) -> str:
        """Return repr representation."""
        return self.name or ""

    def save(self, *args: Any, **kwargs: Any) -> None:
        """Here we generate some default values for IRI."""
        if self.iri is None or self.iri == "":
            self.iri = (
                f"{settings.LARA_PREFIXES['lara']}store/HazardClass/{self.name.replace(' ', '_').lower().strip()}"
            )

        super().save(*args, **kwargs)


class SafetyInformation(models.Model):
    """
    Safety Information Model for substances, polymers, mixtures, ... .

    implements the Safety Information part of the Globally Harmonized System of Classification and Labelling of
    Chemicals (GHS) Standard hazard statements, precautionary statements
    """

    model_curi = "lara:store/SafetyInformation"
    safetyinformation_id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    code = models.TextField(unique=True, help_text="code / unique identifier of the hazard class")
    phrase = models.TextField(blank=True, help_text="phrase of the hazard class")
    hazard_statement = models.BooleanField(default=False, help_text="Is safety hazard statement ?.")
    precautionary_statement = models.BooleanField(default=False, help_text="Is safety precautionary statement ?.")
    iri = models.URLField(
        blank=True,
        null=True,
        unique=True,
        max_length=512,
        help_text="International Resource Identifier - IRI: is used for semantic representation ",
    )
    search_vector = SearchVectorField(null=True)

    class Meta:
        """Meta options for SafetyInformation model."""

        indexes = (GinIndex(fields=["search_vector"]),)

    def __str__(self) -> str:
        """Return string representation."""
        return self.code or ""

    def __repr__(self) -> str:
        """Return repr representation."""
        return self.code or ""

    def save(self, *args: Any, **kwargs: Any) -> None:
        """Here we generate some default values for IRI."""
        if self.iri is None or self.iri == "":
            self.iri = (
                f"{settings.LARA_PREFIXES['lara']}store/SafetyInformation/{self.code.replace(' ', '_').lower().strip()}"
            )

        super().save(*args, **kwargs)


class ItemInstanceAbstr(models.Model):
    """Abstract Item Instance Model fitting for parts, devices, labware, chemicals, organisms, etc."""

    namespace = models.ForeignKey(
        Namespace,
        related_name="%(app_label)s_%(class)s_namespaces_related",
        related_query_name="%(app_label)s_%(class)s_namespaces_related_query",
        on_delete=models.CASCADE,
        null=True,
        blank=True,
        help_text="namespace of item",
    )
    name_display = models.TextField(blank=True, help_text="Human readable name of item, can contain spaces")
    name = models.TextField(
        blank=True,
        help_text="short item name for accessing the item, spaces are replaced by _",
    )
    barcode = models.TextField(
        blank=True,
        unique=True,
        help_text="text representation of a barcode of the item.",
    )
    barcode_json = models.JSONField(
        blank=True,
        null=True,
        help_text="Advanced barcodes in JSON for multiple barcode representations, like QR codes etc.",
    )
    url = models.URLField(blank=True, help_text="Universal Resource Locator - URL")
    pid = models.URLField(blank=True, null=True, unique=True, help_text="Persistent Identifier URI")
    # PAC-ID:  https://github.com/ApiniLabs/PAC-ID
    pac_id = models.CharField(
        blank=True,
        null=True,
        unique=True,
        max_length=255,
        help_text="Publicly Addressable Content IDentifier is a globally unique identifier "
        "that operates independently of a central registry.",
    )
    iri = models.URLField(
        blank=True,
        null=True,
        unique=True,
        help_text="International Resource Identifier - IRI: is used for semantic representation.",
    )
    registration_no = models.TextField(
        blank=True,
        help_text="can be used for internal registrations in a labstore",
    )
    vendor = models.ForeignKey(
        Entity,
        related_name="%(app_label)s_%(class)s_vendors_related",
        related_query_name="%(app_label)s_%(class)s_vendors_related_query",
        on_delete=models.CASCADE,
        blank=True,
        null=True,
        help_text="vendor of the device",
    )
    product_id = models.TextField(blank=True, help_text="vendor product id")
    serial_no = models.TextField(blank=True, help_text="part/devices serial number")
    service_no = models.TextField(blank=True, help_text="service number or tag")
    datetime_manufacturing = models.DateTimeField(null=True, blank=True, help_text="date of manufaction")
    # price - try also https://github.com/django-money/django-money
    price = models.DecimalField(
        blank=True,
        null=True,
        max_digits=14,
        decimal_places=5,
        help_text="price of the part in default currency/EURO",
    )
    currency = models.ForeignKey(
        Currency,
        related_name="%(app_label)s_%(class)s_currencies_related",
        related_query_name="%(app_label)s_%(class)s_currencies_related_query",
        on_delete=models.CASCADE,
        blank=True,
        null=True,
        help_text="currency of the item price",
    )
    charges = models.JSONField(
        blank=True,
        null=True,
        help_text="charges for the item, like toll, but also discounts ...",
    )
    datetime_purchase = models.DateTimeField(null=True, blank=True, help_text="Date of item purchase")
    datetime_delivery = models.DateTimeField(
        blank=True, null=True, help_text="Date of Item delivery = inventarisation date"
    )
    datetime_end_of_warranty = models.DateTimeField(null=True, blank=True, help_text="")
    amount = models.PositiveIntegerField(null=True, blank=True, help_text="amount of items")
    location = models.ForeignKey(
        Location,
        related_name="%(app_label)s_%(class)s_locations_related",
        related_query_name="%(app_label)s_%(class)s_locations_related_query",
        on_delete=models.CASCADE,
        null=True,
        blank=True,
        help_text="location of container",
    )
    storage_info = models.JSONField(
        blank=True,
        null=True,
        help_text="storage information, like temperature, humidity, light,  date opened, date_best_before  ...",
    )
    safety_info = models.JSONField(
        blank=True,
        null=True,
        help_text="safety information, like hazard classes, statements, ...",
    )
    legal_info = models.JSONField(
        blank=True,
        null=True,
        help_text="legal information, like restrictions, licenses, origins of organisms,  ...",
    )
    status = models.ForeignKey(
        ItemStatus,
        related_name="%(app_label)s_%(class)s_status_related",
        related_query_name="%(app_label)s_%(class)s_status_related_query",
        on_delete=models.CASCADE,
        null=True,
        blank=True,
        help_text="status of item",
    )
    users = models.ManyToManyField(
        Entity,
        related_name="%(app_label)s_%(class)s_entities_related",
        related_query_name="%(app_label)s_%(class)s_entities_related_query",
        blank=True,
        help_text="people, who are using and taking care of this item",
    )
    entities_responsible = models.ManyToManyField(
        Entity,
        related_name="%(app_label)s_%(class)s_entities_responsible_related",
        related_query_name="%(app_label)s_%(class)s_entities_responsible_related_query",
        blank=True,
        help_text="entities who are mainly responsible for this item",
    )
    entities_own = models.ManyToManyField(
        Entity,
        related_name="%(app_label)s_%(class)s_entities_own_related",
        related_query_name="%(app_label)s_%(class)s_entities_own_related_query",
        blank=True,
        help_text="entities, who are owning this item",
    )
    groups = models.ManyToManyField(
        Group,
        related_name="%(app_label)s_%(class)s_groups_related",
        related_query_name="%(app_label)s_%(class)s_groups_related_query",
        blank=True,
        help_text="groups of users/entities who can access / and or use this item",
    )
    hash_sha256 = models.CharField(
        max_length=256,
        blank=True,
        help_text="SHA256 hash of item for identity/duplicate checking",
    )
    color = models.TextField(
        blank=True,
        # default=ColorBlind16.full_color_scheme[randint(0, 16)].get_web(),
        help_text="GUI colour representation of the item",
    )
    tags = models.ManyToManyField(
        Tag,
        blank=True,
        related_name="%(app_label)s_%(class)s_tags_related",
        related_query_name="%(app_label)s_%(class)s_tags_related_query",
        help_text="tags",
    )
    # TODO: add role to external resources (as through model)
    resources_external = models.ManyToManyField(
        ExternalResource,
        related_name="%(app_label)s_%(class)s_external_resources_related",
        related_query_name="%(app_label)s_%(class)s_external_resources_related_query",
        blank=True,
        help_text="external resources, like literature, websites, manuals, ...",
    )
    description = models.TextField(blank=True, help_text="description of part instance")
    hardware_config = models.JSONField(
        blank=True,
        null=True,
        help_text="Configuration of the item hardware setup, like connection schemata, wiring, tubing, ...",
    )
    software_config = models.JSONField(
        null=True,
        blank=True,
        help_text="(Internal) Software/Firmware configuration of the item, like interface setup,"
        "ports, DNS, IP, MAC, initialisation parameters, SiLA, OPC-UA configuration, ... ",
    )
    datetime_last_modified = models.DateTimeField(
        blank=True,
        null=True,
        help_text="date and time when data was last modified",  # default=timezone.now,
    )

    search_vector = SearchVectorField(null=True)

    class Meta:
        """Meta options for ItemInstanceAbstr model."""

        abstract = True
        constraints = (
            models.UniqueConstraint(
                fields=["namespace", "name"],
                name="%(app_label)s_%(class)s_namespace_name_unique",
            ),
        )
        indexes = (GinIndex(fields=["search_vector"]),)

    def __str__(self) -> str:  # noqa: D105
        return self.name_full or ""

    def __repr__(self) -> str:  # noqa: D105
        return self.name_full or ""

    def warranty_left(self, day_interval: int = 7) -> float | None:
        """Warranty left in days. Choose day_interval = 7 for weeks."""
        try:
            now = datetime.datetime.now()
            # datetime object is required for subtraction
            warranty_end = datetime.datetime.combine(self.datetime_end_of_warranty, datetime.datetime.min.time())
            warranty_left = warranty_end - now

            return warranty_left.days / day_interval
        except Exception:
            logging.exception("Error in warranty_left")
            return None

    warranty_left.admin_order_field = "datetime_end_of_warranty"  # type: ignore[attr-defined]
    warranty_left.short_description = "Warranty left [weeks]"  # type: ignore[attr-defined]

    def end_of_warranty_soon(self) -> bool | None:  # noqa: D102
        try:
            now = datetime.datetime.now()
            warranty_end = datetime.datetime.combine(self.datetime_end_of_warranty, datetime.datetime.min.time())
            warranty_left = warranty_end - now

            return warranty_left < datetime.timedelta(weeks=12)
        except Exception:
            logging.exception("Error in end_of_warranty_soon")
            return None

    end_of_warranty_soon.admin_order_field = "datetime_end_of_warranty"  # type: ignore[attr-defined]
    end_of_warranty_soon.boolean = True  # type: ignore[attr-defined]
    end_of_warranty_soon.short_description = "Warranty ends soon ?"  # type: ignore[attr-defined]

    def in_warranty_period(self) -> bool:  # noqa: D102
        today = datetime.datetime.now(tz=datetime.UTC)
        return self.datetime_end_of_warranty > today

    in_warranty_period.admin_order_field = "datetime_end_of_warranty"  # type: ignore[attr-defined]
    in_warranty_period.boolean = True  # type: ignore[attr-defined]
    in_warranty_period.short_description = "Still Warranty ?"  # type: ignore[attr-defined]


MAX_QUANTITY = 999999


# validation in django administration
def validate_quantity(value: int) -> None:  # noqa: D103
    if value == 0 or value > MAX_QUANTITY:
        validation_err = f"{value} - is incorrect number, 1-999999 only!"
        raise ValidationError(validation_err)


class OrderItemAbstr(models.Model):
    """
    Abstract class for defining items to be ordered.

    This is required to store quantities, prices, etc. per item
    """

    position = models.IntegerField(null=True, blank=True, help_text="position/order of this item in basket / order")
    quantity = models.PositiveIntegerField(validators=[validate_quantity], default=1)

    item_price_net = models.DecimalField(
        max_digits=10,
        decimal_places=2,
        default=0.0,
        help_text="net price for this item",
    )
    item_price_gross = models.DecimalField(
        blank=True,
        null=True,
        max_digits=10,
        decimal_places=2,
        default=0.0,
        help_text="gross price for this item",
    )
    item_transport_price = models.DecimalField(
        blank=True,
        null=True,
        max_digits=10,
        decimal_places=2,
        default=0.0,
        help_text="transport / shipping price for this item",
    )
    item_price_tax = models.DecimalField(
        blank=True,
        null=True,
        max_digits=10,
        decimal_places=2,
        default=0.0,
        help_text="tax rate for this item",
    )
    item_toll = models.DecimalField(
        blank=True,
        null=True,
        max_digits=10,
        decimal_places=2,
        default=0.0,
        help_text="toll for this item",
    )
    toll_number = models.TextField(blank=True, help_text="description of data type")
    item_discount = models.DecimalField(
        blank=True,
        null=True,
        max_digits=10,
        decimal_places=2,
        default=0.0,
        help_text="discount for this item",
    )
    charges = models.JSONField(
        blank=True,
        null=True,
        help_text="charges for the item, like toll, but also discounts ...",
    )
    total_price_net = models.DecimalField(
        max_digits=10,
        decimal_places=2,
        default=0.0,
        help_text="total price for this item",
    )
    provider = models.ForeignKey(
        Entity,
        related_name="%(app_label)s_%(class)s_order_provider_related",
        related_query_name="%(app_label)s_%(class)s_order_provider_related_query",
        on_delete=models.CASCADE,
        null=True,
        blank=True,
        help_text="provider / company the item was ordered at",
    )
    status = models.ForeignKey(
        ItemStatus,
        related_name="%(app_label)s_%(class)s_orderitem_status_related",
        related_query_name="%(app_label)s_%(class)s_orderitem_status_related_query",
        on_delete=models.CASCADE,
        null=True,
        blank=True,
        help_text="status of the order item - can be used to mark the item as ordered, "
        "requested, quote, arrived delivered, invoiced, paid, ...",
    )

    class Meta:  # noqa: D106
        abstract = True


class ItemOrderAbstr(models.Model):
    """Abstract class for Ordering Items (OrderItems)."""

    namespace = models.ForeignKey(
        Namespace,
        related_name="%(app_label)s_%(class)s_namespaces_related",
        related_query_name="%(app_label)s_%(class)s_namespaces_related_query",
        on_delete=models.CASCADE,
        null=True,
        blank=True,
        help_text="namespace of order ",
    )
    name_display = models.TextField(
        blank=True,
        help_text="short, human readable title of order, can contain whitespaces",
    )
    name = models.TextField(
        blank=True,
        help_text="short name of order without whitespaces and special characters",
    )
    order_id_internal = models.TextField(blank=True, help_text="internal id / number of order")
    order_id_external = models.TextField(blank=True, help_text="external id / number of order")
    order_barcode = models.TextField(
        blank=True,
        unique=True,
        help_text="barcode of order, can be, e.g., used for tracking or in-house delivery.",
    )
    order_barcode_json = models.JSONField(
        blank=True,
        null=True,
        help_text="Advanced barcodes in JSON for multiple barcode representations, like QR codes etc.",
    )
    iri = models.URLField(
        blank=True,
        null=True,
        unique=True,
        max_length=512,
        help_text="International Resource Identifier - IRI: is used for semantic representation ",
    )
    datetime_order = models.DateTimeField(blank=True, null=True, help_text="date of order")

    entity_ordered = models.ForeignKey(
        Entity,
        related_name="%(app_label)s_%(class)s_entities_ordered_related",
        related_query_name="%(app_label)s_%(class)s_entities_ordered_related_query",
        on_delete=models.CASCADE,
        null=True,
        blank=True,
        help_text="entity who ordered the item",
    )
    group_ordered = models.ForeignKey(
        Group,
        related_name="%(app_label)s_%(class)s_groups_ordered_related",
        related_query_name="%(app_label)s_%(class)s_groups_ordered_related_query",
        on_delete=models.CASCADE,
        null=True,
        blank=True,
        help_text="group who ordered the item",
    )
    entity_ordering = models.ForeignKey(
        Entity,
        related_name="%(app_label)s_%(class)s_entities_ordering_related",
        related_query_name="%(app_label)s_%(class)s_entities_ordering_related_query",
        on_delete=models.CASCADE,
        null=True,
        blank=True,
        help_text="entity who is ordering the item (from an external resource like a company)",
    )
    group_ordering = models.ForeignKey(
        Group,
        related_name="%(app_label)s_%(class)s_groups_ordering_related",
        related_query_name="%(app_label)s_%(class)s_groups_ordering_related_query",
        on_delete=models.CASCADE,
        null=True,
        blank=True,
        help_text="group who is ordering the item (from an external resource like a company)",
    )
    provider = models.ForeignKey(
        Entity,
        related_name="%(app_label)s_%(class)s_order_provider_related",
        related_query_name="%(app_label)s_%(class)s_order_provider_related_query",
        on_delete=models.CASCADE,
        null=True,
        blank=True,
        help_text="provider / company providing the order.",
    )
    order_quote_id = models.TextField(blank=True, help_text="quote id of order")
    datetime_order_quote = models.DateTimeField(blank=True, null=True, help_text="date of latest quote")
    order_invoice_id = models.TextField(blank=True, help_text="invoice id of order")
    datetime_order_invoice = models.DateTimeField(blank=True, null=True, help_text="date of latest invoice")
    # transport / shipping
    transport_company = models.ForeignKey(
        Entity,
        related_name="%(app_label)s_%(class)s_transport_companies_related",
        related_query_name="%(app_label)s_%(class)s_transport_companies_related_query",
        on_delete=models.CASCADE,
        null=True,
        blank=True,
        help_text="company transport the item",
    )
    transport_tracking_id = models.TextField(blank=True, help_text="tracking ID of transport.")
    transport_price = models.DecimalField(
        max_digits=10,
        decimal_places=2,
        blank=True,
        null=True,
        help_text="price of transport / shipping",
    )
    # add budget
    budget_url = models.URLField(blank=True, max_length=512, help_text="Link to budget")
    datetime_order_pay = models.DateTimeField(
        default=timezone.now, blank=True, null=True, help_text="date of order payment"
    )

    @property
    def order_total_price_net(self) -> float:
        """Calculate the total price of the order net."""
        return sum([item.total_price_net for item in self.deviceorderitem_set.all()])

    order_total_price_gross = models.DecimalField(
        max_digits=10,
        decimal_places=2,
        blank=True,
        null=True,
        help_text="total price of order gross",
    )

    order_total_price_tax = models.DecimalField(
        max_digits=10,
        decimal_places=2,
        blank=True,
        null=True,
        help_text="total price of order TAX/VAT",
    )

    order_total_price_toll = models.DecimalField(
        max_digits=10,
        decimal_places=2,
        blank=True,
        null=True,
        help_text="total price of order toll",
    )
    order_currency = models.ForeignKey(
        Currency,
        related_name="%(app_label)s_%(class)s_order_total_price_currencies_related",
        related_query_name="%(app_label)s_%(class)s_order_total_price_currencies_related_query",
        on_delete=models.CASCADE,
        null=True,
        blank=True,
        help_text="currency of order total price",
    )
    order_status = models.ForeignKey(
        ItemStatus,
        related_name="%(app_label)s_%(class)s_order_status_related",
        related_query_name="%(app_label)s_%(class)s_order_status_related_query",
        on_delete=models.CASCADE,
        null=True,
        blank=True,
        help_text="state of the order, like ordered, requested, quote, arrived delivered, invoiced, paid, ...",
    )
    charges = models.JSONField(
        blank=True,
        null=True,
        help_text="charges for the order, like toll, but also discounts ...",
    )
    datetime_last_modified = models.DateTimeField(
        default=timezone.now,
        blank=True,
        null=True,
        help_text="datetime of last changes of the order",
    )

    order_quotes = models.ManyToManyField(
        "ExtraData",  # String reference - will be resolved per app
        related_name="%(app_label)s_%(class)s_order_quotes",
        blank=True,
        help_text="Information about quotes for this order, like PDFs, Images, URLs, ...",
    )
    order_invoices = models.ManyToManyField(
        "ExtraData",  # String reference - will be resolved per app
        related_name="%(app_label)s_%(class)s_order_invoices",
        blank=True,
        help_text="Information about invoices for this order, like PDFs, Images, URLs, ...",
    )
    data_extra = models.ManyToManyField(
        "ExtraData",  # String reference - will be resolved per app
        related_name="%(app_label)s_%(class)s_data_extra",
        blank=True,
        help_text="additional data for this order",
    )

    search_vector = SearchVectorField(null=True)

    class Meta:  # noqa: D106
        abstract = True
        constraints = (
            models.UniqueConstraint(
                fields=["namespace", "name"],
                name="%(app_label)s_%(class)s_namespace_name_unique",
            ),
        )
        indexes = (GinIndex(fields=["search_vector"]),)

    def __str__(self) -> str:  # noqa: D105
        return self.name or ""

    def __repr__(self) -> str:  # noqa: D105
        return self.name or ""

    def save(self, *args: Any, **kwargs: Any) -> None:  # noqa: D102
        if self.order_barcode == "":
            self.order_barcode = f"{self.name}_{self.pk}"
        super().save(*args, **kwargs)


class ItemBasketAbstr(models.Model):
    """
    Abstract class for collecting OrderItems in a basket or wishlist for (re-)ordering.

    :note: an item stash/local store can be generated by subclassing ItemBasketAbstr class and
           adding a location
    """

    namespace = models.ForeignKey(
        Namespace,
        related_name="%(app_label)s_%(class)s_namespaces_related",
        related_query_name="%(app_label)s_%(class)s_namespaces_related_query",
        on_delete=models.CASCADE,
        null=True,
        blank=True,
        help_text="namespace of the basket",
    )
    name_display = models.TextField(blank=True, help_text="Human readable name of basket, can contain spaces")
    name = models.TextField(
        unique=True,
        blank=True,
        help_text="Name of the basket without spaces and special characters, for accessing the basket",
    )
    entity = models.ForeignKey(
        Entity,
        related_name="%(app_label)s_%(class)s_entities_related",
        related_query_name="%(app_label)s_%(class)s_entities_related_query",
        on_delete=models.CASCADE,
        null=True,
        blank=True,
        help_text="entity owning the basket",
    )
    datetime_created = models.DateTimeField(
        default=timezone.now,
        blank=True,
        null=True,
        help_text="datetime when the basket was created",
    )
    datetime_last_modified = models.DateTimeField(
        default=timezone.now,
        blank=True,
        null=True,
        help_text="datetime of last changes of the basket",
    )
    search_vector = SearchVectorField(null=True)

    class Meta:  # noqa: D106
        abstract = True
        indexes = (GinIndex(fields=["search_vector"]),)

    def add_item_instance(self, item_instance: Any, quantity: int = 1) -> None:
        """Standardized method name for adding items to baskets."""
        raise NotImplementedError("Subclasses must implement add_item_instance")



"""
A custom manager for calculating the total price of items in a shopping cart.

Additional managers for calculating the budget for a given project. It also includes
a custom queryset for filtering items based on the user's entity.

s. https://www.youtube.com/watch?v=n47VFgqG-E8&t=1463s
"""

from django.db import models
from django.db.models import Sum, F, FloatField
from django.db.models.functions import Coalesce


class BasketManager(models.Manager):
    """Custom manager for Basket model to calculate total price and budget."""

    def total_price(self, basket_id):
        """Calculate the total price of items in the basket."""
        return (
            self.filter(pk=basket_id)
            .annotate(total_price=Coalesce(Sum(F("items__quantity") * F("items__price"), output_field=FloatField()), 0))
            .values_list("total_price", flat=True)
            .first()
        )

    def budget_for_project(self, project_id):
        """Calculate the total budget for a given project."""
        return (
            self.filter(project_id=project_id)
            .annotate(
                total_budget=Coalesce(Sum(F("items__quantity") * F("items__price"), output_field=FloatField()), 0)
            )
            .values_list("total_budget", flat=True)
            .first()
        )


class ItemQuerySet(models.QuerySet):
    """Custom queryset for filtering items based on user's entity."""

    def for_user(self, user):
        """Filter items based on the user's entity."""
        return self.filter(basket__entity=user.entity)


class ItemManager(models.Manager):
    """Custom manager for Item model to use the custom queryset."""

    def get_queryset(self):
        """Return the custom queryset."""
        return ItemQuerySet(self.model, using=self._db)

    def for_user(self, user):
        """Filter items based on the user's entity."""
        return self.get_queryset().for_user(user)


class OrderItemManager(models.Manager):
    """Custom manager for OrderItem model to use the custom queryset."""

    def get_queryset(self):
        """Return the custom queryset."""
        return ItemQuerySet(self.model, using=self._db)

    def for_user(self, user):
        """Filter order items based on the user's entity."""
        return self.get_queryset().for_user(user)

    def total_price_for_user(self, user):
        """Calculate the total price of order items for a given user."""
        return (
            self.for_user(user)
            .annotate(total_price=Coalesce(Sum(F("quantity") * F("price"), output_field=FloatField()), 0))
            .values_list("total_price", flat=True)
            .first()
        )


class ProjectManager(models.Manager):
    """Custom manager for Project model to calculate total budget."""

    def total_budget_for_user(self, user):
        """Calculate the total budget for a given user's projects."""
        return (
            self.filter(entity=user.entity)
            .annotate(
                total_budget=Coalesce(
                    Sum(F("baskets__items__quantity") * F("baskets__items__price"), output_field=FloatField()), 0
                )
            )
            .values_list("total_budget", flat=True)
            .first()
        )



# ruff: noqa: D400, D415
"""
_____________________________________________________________________

:PROJECT: LARAsuite

*lara_django_store forms.*

:details: lara_django_store admin module admin backend configuration.
"""



# ruff: noqa: D400, D415
"""
_____________________________________________________________________

:PROJECT: LARAsuite

*lara_django_store app.*

:details: lara_django_store app configuration.
         This provides a generic django app configuration mechanism.
         For more details see:
         https://docs.djangoproject.com/en/4.0/ref/applications/
:authors: mark doerr <mark.doerr@uni-greifswald.de>

.. note:: -
.. todo:: -
________________________________________________________________________
"""

from django.apps import AppConfig


class LaraStoreConfig(AppConfig):
    """Configuration for the lara_django_store Django app."""

    name = "lara_django_store"
    # url_path = 'store'
    # enter a verbose name for your app: lara_django_store here - this will be used in the admin interface
    verbose_name = "LARA-django Store"
    # lara_app_icon = 'lara_django_store_icon.svg'  # this will be used to display an icon, e.g. in the main LARA menu.
